"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const human_list_1 = require("human-list");
const alphabetLower = 'abcdefghijklmnopqrstuvwxyz';
const alphabetUpper = alphabetLower.toUpperCase();
const alphabetAll = alphabetLower + alphabetUpper;
const digits = '0123456789';
const alphanumeric = alphabetAll + digits;
const symbols = `!@#$%^&*()_-+=[{]}\\|><.,?/"';:~\``;
const charsToHash = chars => chars
    .split('')
    .reduce((prev, cur) => {
    prev[cur] = true;
    return prev;
}, {});
exports.charHashes = {
    letters: charsToHash(alphabetAll),
    lowercase: charsToHash(alphabetLower),
    uppercase: charsToHash(alphabetUpper),
    digits: charsToHash(digits),
    symbols: charsToHash(symbols),
    alphanumeric: charsToHash(alphanumeric),
    spaces: charsToHash(" ")
};
exports.getHash = str => exports.charHashes[str] ? exports.charHashes[str] : charsToHash(str);
exports.combineHashes = (...hashes) => hashes.reduce((prev, cur) => ({ ...prev, ...exports.getHash(cur) }), {});
const getHumanList = (arr, conjunction = "and") => [arr.slice(0, -1).join(', '), arr.slice(-1)[0]]
    .join(arr.length < 2 ? '' : arr.length < 3 ? ` ${conjunction} ` : `, ${conjunction} `);
exports.createBinaryHash = (acc, cur, idx) => {
    const binaryRepresentation = (1 << idx);
    for (const char in cur) {
        acc[char] = (acc[char] || 0) | binaryRepresentation;
    }
    return acc;
};
exports.surroundQuotes = str => `"${str}"`;
exports.createMustContainValidator = (mustContain) => {
    const mustContainNames = [].concat(mustContain)
        .map(name => exports.charHashes[name] ? camelToLower(name) : name);
    const binaryHash = mustContainNames
        .reverse()
        .map(exports.getHash)
        .reduce(exports.createBinaryHash, {});
    const checkVal = parseInt("1".repeat(mustContain.length), 2);
    return str => {
        let curVal = 0;
        for (const char of str) {
            if ((curVal |= binaryHash[char]) === checkVal) {
                return [];
            }
        }
        // not found, so retrieve names of missing values
        const notContained = [];
        for (const idx in mustContainNames) {
            if (!(curVal & 1)) {
                notContained.push(mustContainNames[idx]);
            }
            curVal >>= 1;
        }
        const humanList = getHumanList(notContained.map(exports.surroundQuotes));
        return [new Error(`Must contain ${humanList}`)];
    };
};
exports.createStringMinValidator = (min) => str => {
    if (str.length < min) {
        return [new Error(`Must be at least ${min} characters long.`)];
    }
};
exports.createStringMaxValidator = (max) => str => {
    if (str.length > max) {
        return [new Error(`Must be no longer than ${max} characters long.`)];
    }
};
exports.createValidCharsValidator = (validChars) => {
    const validCharsArr = [].concat(validChars);
    const humanList = getHumanList(validCharsArr.map(exports.surroundQuotes));
    const errorMessageTail = ` is not allowed. Only ${humanList} are allowed.`;
    const validCharsCombined = exports.combineHashes(...validCharsArr);
    return str => {
        for (const char of str) {
            if (!validCharsCombined[char]) {
                return [new Error(char + errorMessageTail)];
            }
        }
        return [];
    };
};
exports.createRegexpValidator = (regexp) => str => {
    if (!regexp.test(str)) {
        return [new Error(`Invalid input.`)];
    }
};
const defaultValidateOptions = {
    maxErrors: Infinity,
};
exports.createStringValidator = (requirements) => {
    const validatorFns = [];
    if (requirements.mustContain) {
        validatorFns.push(exports.createMustContainValidator(requirements.mustContain));
    }
    if (requirements.validChars) {
        validatorFns.push(exports.createValidCharsValidator(requirements.validChars));
    }
    if (requirements.min) {
        validatorFns.push(exports.createStringMinValidator(requirements.min));
    }
    if (requirements.max) {
        validatorFns.push(exports.createStringMaxValidator(requirements.max));
    }
    if (requirements.regexp) {
        const regexp = requirements.regexp instanceof RegExp ? requirements.regexp : new RegExp(requirements.regexp);
        validatorFns.push(exports.createRegexpValidator(regexp));
    }
    return (str, options = defaultValidateOptions) => {
        let errors = [];
        for (const validatorFn of validatorFns) {
            const newError = validatorFn(str);
            if (newError) {
                errors = errors.concat(newError);
                if (errors.length > options.maxErrors) {
                    break;
                }
            }
        }
        return errors;
    };
};
exports.createNumberValidator = ({ min, max }) => {
    max = max || Infinity;
    min = min || -Infinity;
    return (num) => {
        if (typeof num !== "number") {
            const originalNumber = num;
            num = Number(num);
            if (Number.isNaN(num)) {
                return [new Error(`Could not convert ${originalNumber} to a number.`)];
            }
        }
        if (num < min) {
            return [new Error(`Must be greater than or equal to ${min}.`)];
        }
        if (num > max) {
            return [new Error(`Must be less than or equal to ${max}.`)];
        }
        return [];
    };
};
exports.createOneOfValidator = ({ options }) => {
    const err = [new Error(`Must be ${human_list_1.orList(options)}.`)];
    return input => {
        if (!options.includes(input)) {
            return err;
        }
        return [];
    };
};
function camelToLower(camelCase) {
    return camelCase
        .replace(/([A-Z])/g, (match) => ` ${match}`)
        .replace(/ ./g, (match) => match.toLowerCase())
        .trim();
}
//# sourceMappingURL=validate.js.map